# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Pipeline I/O using the local filesystem as a storage backend.

Note that pipeline processing *always* uses the local filesystem for
intermediate steps. This module is for scenarios where the final long-term
storage of processed data should also involve the local filesystem.

"""

__all__ = '''
LocalPipelineIo
'''.split()

import os.path
import shutil

from . import PipelineIo


class LocalPipelineIo(PipelineIo):
    """
    I/O for pipeline processing using the local disk.

    Parameters
    ----------
    path_prefix : str
        A path prefix that will be used for all I/O options.

    """
    _path_prefix = None

    def __init__(self, path_prefix):
        self._path_prefix = path_prefix

    def _export_config(self):
        return {
            '_type': 'local',
            'path': self._path_prefix,
        }

    @classmethod
    def _new_from_config(cls, config):
        return cls(config['path'])

    def _make_item_name(self, path_array):
        return os.path.join(self._path_prefix, *path_array)

    def check_exists(self, *path):
        return os.path.exists(self._make_item_name(path))

    def get_item(self, *path, dest=None):
        with open(self._make_item_name(path), 'rb') as f:
            shutil.copyfileobj(f, dest)

    def put_item(self, *path, source=None):
        fpath = self._make_item_name(path)

        cdir = os.path.split(fpath)[0]
        os.makedirs(cdir, exist_ok=True)

        with open(fpath, 'wb') as f:
            shutil.copyfileobj(source, f)

    def list_items(self, *path):
        dpath = self._make_item_name(path)

        for stem in os.listdir(dpath):
            yield stem, os.path.isdir(os.path.join(dpath, stem))



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

from __future__ import absolute_import, division, print_function

import numpy as np
import numpy.testing as nt
import os.path
import pytest
import sys
from xml.etree import ElementTree as etree

from . import assert_xml_elements_equal, test_path
from ..builder import Builder
from .. import cli
from .. import collection
from .. import multi_tan


try:
    from astropy.io import fits
    HAS_ASTRO = True
except ImportError:
    HAS_ASTRO = False


try:
    import reproject
    HAS_REPROJECT = True
except ImportError:
    HAS_REPROJECT = False


class TestMultiTan(object):
    WTML = """
<Folder Browseable="True" Group="Explorer" Name="Toasty" MSRCommunityId="0"
        MSRComponentId="0" Permission="0" Searchable="True" Type="Sky">
  <Place Angle="0.0" AngularSize="0.0" DataSetType="Sky" Dec="0.7438249862258411"
         Distance="0.0" DomeAlt="0.0" DomeAz="0.0" Lat="0.0" Lng="0.0"
         Magnitude="0.0" MSRCommunityId="0" MSRComponentId="0" Name="Toasty"
         Opacity="100.0" Permission="0" RA="14.41975153073335" Rotation="0.0"
         Thumbnail="thumb.jpg" ZoomLevel="0.2437119999998555" >
    <ForegroundImageSet>
      <ImageSet BandPass="Visible" BaseDegreesPerTile="0.023893333333319167"
                BaseTileLevel="0" BottomsUp="False" CenterX="216.2962962963"
                CenterY="0.74380165289257" DataSetType="Sky" ElevationModel="False"
                FileType=".fits" Generic="False" MeanRadius="0.0" MSRCommunityId="0"
                MSRComponentId="0" Name="Toasty" OffsetX="2.33333333333195e-05"
                OffsetY="2.33333333333195e-05" Permission="0" Projection="Tan"
                Rotation="-0.0" Sparse="True" StockSet="False" TileLevels="1"
                Url="{1}/{3}/{3}_{2}.fits" WidthFactor="2">
        <ThumbnailUrl>thumb.jpg</ThumbnailUrl>
      </ImageSet>
    </ForegroundImageSet>
  </Place>
</Folder>"""

    # Gross workaround for platform differences in the XML output.

    if sys.platform == 'darwin':
        WTML = WTML.replace('Dec="0.7438249862258411"', 'Dec="0.743824986225841"')

    # Back to the non-gross stuff.

    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)


    def test_basic(self):
        coll = collection.SimpleFitsCollection([test_path('wcs512.fits.gz')])

        proc = multi_tan.MultiTanProcessor(coll)

        from ..pyramid import PyramidIO
        pio = PyramidIO(self.work_path('basic'), default_format='fits')

        builder = Builder(pio)

        proc.compute_global_pixelization(builder)
        proc.tile(pio)

    BARY_SLICES = [
        (slice(0, 128), slice(0, 128)),
        (slice(0, 128), slice(128, None)),
        (slice(128, None), slice(0, 128)),
        (slice(128, None), slice(128, None)),
    ]

    def maybe_test_barycenter(self, path, bary_expected):
        """
        Check the barycenters of four 128x128 quadrants of a tile file. The idea
        here is that if we introduce a problem with vertical flips in tiled FITS
        processing, we'll detect it here.
        """

        if not HAS_ASTRO:
            return

        with fits.open(path) as hdul:
            data = hdul[0].data

        data[~np.isfinite(data)] = 0.0
        bary_observed = []

        for islice in self.BARY_SLICES:
            idata = data[islice]
            yidx, xidx = np.indices((128, 128))
            xbary = (idata * xidx).sum() / idata.sum()
            ybary = (idata * yidx).sum() / idata.sum()
            bary_observed.append((xbary, ybary))

        nt.assert_array_almost_equal(bary_observed, bary_expected, decimal=5)

    WCS512_BARYDATA = [
        (63.44949378800272, 64.40535387506924),
        (63.24744175084746, 63.67473452789256),
        (65.22950207855361, 63.35629429568745),
        (62.027396724898814, 62.815937534782144)
    ]

    def test_basic_cli(self):
        """
        Test the CLI interface. We don't go out of our way to validate the
        computations in detail -- that's for the unit tests that probe the
        module directly.
        """
        expected = etree.fromstring(
            self.WTML
                .replace('Thumbnail="thumb.jpg"', '')
                .replace('<ThumbnailUrl>thumb.jpg</ThumbnailUrl>', '')
        )

        args = [
            'tile-multi-tan',
            '--hdu-index', '0',
            '--outdir', self.work_path('basic_cli'),
            test_path('wcs512.fits.gz')
        ]
        cli.entrypoint(args)

        with open(self.work_path('basic_cli', 'index_rel.wtml'), 'rt', encoding='utf8') as f:
            observed = etree.fromstring(f.read())

        assert_xml_elements_equal(observed, expected)

        args = [
            'cascade',
            '--start', '1',
            self.work_path('basic_cli'),
        ]
        cli.entrypoint(args)

        self.maybe_test_barycenter(self.work_path('basic_cli', '0', '0', '0_0.fits'), self.WCS512_BARYDATA)


    def test_study_cli(self):
        """
        Test tile-study on FITS. This should properly go in test_study.py, but
        this file is the one that has the reference WTML information.
        """
        expected = etree.fromstring(self.WTML)

        args = [
            'tile-study',
            '--placeholder-thumbnail',
            '--outdir', self.work_path('study_cli'),
            test_path('wcs512.fits.gz')
        ]
        cli.entrypoint(args)

        with open(self.work_path('study_cli', 'index_rel.wtml'), 'rt', encoding='utf8') as f:
            observed = etree.fromstring(f.read())

        assert_xml_elements_equal(observed, expected)

        args = [
            'cascade',
            '--start', '1',
            self.work_path('study_cli'),
        ]
        cli.entrypoint(args)

        self.maybe_test_barycenter(self.work_path('study_cli', '0', '0', '0_0.fits'), self.WCS512_BARYDATA)


    @pytest.mark.skipif('not HAS_REPROJECT')
    def test_as_multi_wcs(self):
        """
        Once again, this doesn't super belong here, but this is where we have
        the reference data. We don't compare the WTML contents here since the
        reprojection isn't going to preserve the WCS in detail.
        """
        from .. import builder, collection, multi_wcs, pyramid

        reproject_function = reproject.reproject_interp
        outdir = self.work_path('as_multi_wcs')

        pio = pyramid.PyramidIO(outdir, default_format='fits')
        bld = builder.Builder(pio)
        coll = collection.SimpleFitsCollection([test_path('wcs512.fits.gz')], hdu_index=0)
        proc = multi_wcs.MultiWcsProcessor(coll)
        proc.compute_global_pixelization(bld)
        proc.tile(pio, reproject_function, cli_progress=False, parallel=1)
        bld.write_index_rel_wtml()

        args = [
            'cascade',
            '--start', '1',
            self.work_path('as_multi_wcs'),
        ]
        cli.entrypoint(args)

        self.maybe_test_barycenter(self.work_path('as_multi_wcs', '0', '0', '0_0.fits'), self.WCS512_BARYDATA)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2019 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""General tools for working with tile pyramids.

Toasty and the AAS WorldWide Telescope support two kinds of tile pyramid
formats: the all-sky TOAST projection, and “studies” which are tile pyramids
rooted in a subset of the sky using a tangential projection. Both kinds of
tile pyramids have much in common, and this module implements their
overlapping functionality.

"""
from __future__ import absolute_import, division, print_function

__all__ = '''
depth2tiles
generate_pos
is_subtile
next_highest_power_of_2
Pos
pos_children
pos_parent
PyramidIO
tiles_at_depth
'''.split()

import glob
from collections import namedtuple
from contextlib import contextmanager
import numpy as np
import os.path

from .image import ImageLoader, SUPPORTED_FORMATS, get_format_vertical_parity_sign

Pos = namedtuple('Pos', 'n x y')


def next_highest_power_of_2(n):
    """Ugh, this implementation is so dumb.

    We also assume that we are being called in a tiling context, in which case
    numbers less than 256 should be bumped up to 256 (the number of pixels in
    a single tile).

    """
    p = 256
    while p < n:
        p *= 2
    return p


def depth2tiles(depth):
    """Return the total number of tiles in a WWT tile pyramid of depth *depth*."""
    return (4 ** (depth + 1) - 1) // 3


def tiles_at_depth(depth):
    """
    Return the number of tiles in the WWT tile pyramid layer at depth *depth*.
    """
    return 4**depth


def is_subtile(deeper_pos, shallower_pos):
    """Determine if one tile is a child of another.

    Parameters
    ----------
    deeper_pos : Pos
        A tile position.
    shallower_pos : Pos
        A tile position that is shallower than *deeper_pos*.

    Returns
    -------
    True if *deeper_pos* represents a tile that is a child of *shallower_pos*.

    """
    if deeper_pos.n < shallower_pos.n:
        raise ValueError('deeper_pos has a lower depth than shallower_pos')

    if deeper_pos.n == shallower_pos.n:
        return deeper_pos.x == shallower_pos.x and deeper_pos.y == shallower_pos.y

    return is_subtile(pos_parent(deeper_pos)[0], shallower_pos)


def pos_parent(pos):
    """Return a tile position's parent.

    Parameters
    ----------
    pos : Pos
        A tile position.

    Returns
    -------
    parent : Pos
        The tile position that is the parent of *pos*.
    x_index : integer, 0 or 1
        The horizontal index of the child inside its parent.
    y_index : integer, 0 or 1
        The vertical index of the child inside its parent.

    """
    if pos.n < 1:
        raise ValueError('cannot take the parent of a tile position with depth < 1')

    parent = Pos(
        n = pos.n - 1,
        x = pos.x // 2,
        y = pos.y // 2
    )
    return parent, pos.x % 2, pos.y % 2


def pos_children(pos):
    """Return the children of a tile position.

    Parameters
    ----------
    pos : :class:`Pos`
        A tile position.

    Returns
    -------
    A list of four child :class:`Pos` instances. The return value is
    guaranteed to always be a list, and the order of the children will always
    be: top left, top right, bottom left, bottom right.

    """
    n, x, y = pos.n, pos.x, pos.y
    n += 1
    x *= 2
    y *= 2

    return [
        Pos(n=n, x=x,     y=y    ),
        Pos(n=n, x=x + 1, y=y    ),
        Pos(n=n, x=x,     y=y + 1),
        Pos(n=n, x=x + 1, y=y + 1),
    ]


def _postfix_pos(pos, depth):
    if pos.n > depth:
        return

    for immed_child in pos_children(pos):
        for item in _postfix_pos(immed_child, depth):
            yield item

    yield pos


def generate_pos(depth):
    """Generate a pyramid of tile positions.

    The generate proceeds in a broadly deeper-first fashion. In particular, if
    a position *p* is yielded, you can assume that its four children have been
    yielded previously, unless the depth of *p* is equal to *depth*.

    Parameters
    ----------
    depth : int
        The tile depth to recurse to.

    Yields
    ------
    pos : :class:`Pos`
        An individual position to process.

    """
    for item in _postfix_pos(Pos(0, 0, 0), depth):
        yield item


class PyramidIO(object):
    """
    Manage I/O on a tile pyramid.

    Parameters
    ----------
    base_dir : str
        The base directory containing the tiles
    scheme : str
        The tile organization scheme, should be either 'L/Y/YX' or 'LXY'
    default_format : str
        The file format to assume for the tiles if none is specified when
        reading/writing tiles. If not specified, and base_dir exists and
        contains files, these are used to guess default_format. Otherwise
        defaults to 'png'.
    """

    def __init__(self, base_dir, scheme='L/Y/YX', default_format=None):

        self._base_dir = base_dir

        if scheme == 'L/Y/YX':
            self._tile_path = self._tile_path_LsYsYX
            self._scheme = '{1}/{3}/{3}_{2}'
            tile_pattern = "*/*/*_*.*"
        elif scheme == 'LXY':
            self._tile_path = self._tile_path_LXY
            self._scheme = 'L{1}X{2}Y{3}'
            tile_pattern = "L*X*Y*.*"
        else:
            raise ValueError(f'unsupported "scheme" option for PyramidIO: {scheme}')

        if default_format is None:
            default_format = 'png'

            if os.path.exists(base_dir) and os.path.isdir(base_dir):
                for filename in glob.iglob(os.path.join(base_dir, tile_pattern)):
                    extension = os.path.splitext(filename)[1][1:]
                    if extension in SUPPORTED_FORMATS:
                        default_format = extension
                        break

        self._default_format = default_format

    def tile_path(self, pos, format=None):
        """Get the path for a tile, creating its containing directories.

        Parameters
        ----------
        pos : Pos
            The tile to get a path for.test_plate_carree_ecliptic

        Returns
        -------
        The path as a string.

        Notes
        -----
        This function does I/O itself — it creates the parent directories
        containing the tile path. It is not an error for the parent
        directories to already exist.

        """
        level = str(pos.n)
        ix = str(pos.x)
        iy = str(pos.y)
        return self._tile_path(level, ix, iy, format=format)

    def _tile_path_LsYsYX(self, level, ix, iy, format=None):
        d = os.path.join(self._base_dir, level, iy)

        # We can't use the `exist_ok` kwarg because it's not available in Python 2.
        try:
            os.makedirs(d)
        except OSError as e:
            if e.errno != 17:
                raise  # not EEXIST

        return os.path.join(d, '{}_{}.{}'.format(iy, ix, format or self._default_format))

    def _tile_path_LXY(self, level, ix, iy, format=None):
        # We can't use the `exist_ok` kwarg because it's not available in Python 2.
        try:
            os.makedirs(self._base_dir)
        except OSError as e:
            if e.errno != 17:
                raise  # not EEXIST

        return os.path.join(
            self._base_dir,
            'L{}X{}Y{}.{}'.format(level, ix, iy, format or self._default_format)
        )

    def get_path_scheme(self):
        """Get the scheme for buiding tile paths as used in the WTML standard.

        Returns
        -------
        The naming scheme, a string resembling ``{1}/{3}/{3}_{2}``.

        Notes
        -----
        The naming scheme is currently hardcoded to be the format given above,
        but in the future other options might become available.

        """
        return self._scheme

    def get_default_format(self):
        """
        Get the default image storage format for this pyramid.

        Returns
        -------
        The format, a string resembling ``"png"`` or ``"fits"``.
        """
        return self._default_format

    def get_default_vertical_parity_sign(self):
        """
        Get the parity sign (vertical data layout) of the tiles in this pyramid's
        default format.

        Returns
        -------
        Either +1 or -1, depending on the format.
        """

        if self._default_format is None:
            raise Exception('cannot get default parity sign without a default format')
        return get_format_vertical_parity_sign(self._default_format)

    def read_image(self, pos, default='none', masked_mode=None, format=None):
        """
        Read an Image for the specified tile position.

        Parameters
        ----------
        pos : :class:`Pos`
            The tile position to read.
        default : str, defaults to "none"
            What to do if the specified tile file does not exist. If this is
            "none", ``None`` will be returned instead of an image. If this is
            "masked", an all-masked image will be returned, using
            :meth:`~toasty.image.ImageMode.make_maskable_buffer`.
            Otherwise, :exc:`ValueError` will be raised.
        masked_mode : :class:`toasty.image.ImageMode`
            The image data mode to use if ``default`` is set to ``'masked'``.
        """
        p = self.tile_path(pos, format=format)

        loader = ImageLoader()

        try:
            img = loader.load_path(p)
        except IOError as e:
            if e.errno != 2:
                raise  # not EEXIST

            if default == 'none':
                return None
            elif default == 'masked':
                if masked_mode is None:
                    raise ValueError('masked_mode should be set if default="masked"')
                buf = masked_mode.make_maskable_buffer(256, 256)
                buf.clear()
                return buf
            else:
                raise ValueError('unexpected value for "default": {!r}'.format(default))

        return img

    def write_image(self, pos, image, format=None, mode=None):
        """Write an Image for the specified tile position.

        Parameters
        ----------
        pos : :class:`Pos`
            The tile position to write.
        image : :class:`toasty.image.Image`
            The image to write.

        """
        p = self.tile_path(pos, format=format or self._default_format)
        image.save(p, format=format or self._default_format, mode=mode)

    @contextmanager
    def update_image(self, pos, default='none', masked_mode=None, format=None):
        from filelock import FileLock

        p = self.tile_path(pos)

        with FileLock(p + '.lock'):
            img = self.read_image(
                pos,
                default=default,
                masked_mode=masked_mode,
                format=format or self._default_format
            )

            yield img
            self.write_image(pos, img, format=format or self._default_format)

    def clean_lockfiles(self, level):
        """
        Clean up any lockfiles created during parallelized pyramid creation.

        Parameters
        ----------
        level : :class:`int` A tile pyramid depth.

        Notes
        -----
        If you use the :meth:`update_image` method, you should call this
        function after your processing is complete to remove lockfiles that are
        created to ensure that multiple processes don't stomp on each other's
        work. The "cascade" stage doesn't need locking, so in general only the
        deepest level of the pyramid will need to be cleaned.
        """
        for x in range(0, 2**level):
            for y in range(0, 2**level):
                pos = Pos(level, x, y)
                p = self.tile_path(pos) + '.lock'

                try:
                    os.unlink(p)
                except FileNotFoundError:
                    pass

    def open_metadata_for_read(self, basename):
        """
        Open a metadata file in read mode.

        Parameters
        ----------
        basename : str
            The basename of the metadata file

        Returns
        -------
        A readable and closeable file-like object returning bytes.

        """
        return open(os.path.join(self._base_dir, basename), 'rb')

    def open_metadata_for_write(self, basename):
        """
        Open a metadata file in write mode.

        Parameters
        ----------
        basename : str
            The basename of the metadata file

        Returns
        -------
        A writable and closeable file-like object accepting bytes.

        """
        # We can't use the `exist_ok` kwarg because it's not available in Python 2.
        try:
            os.makedirs(self._base_dir)
        except OSError as e:
            if e.errno != 17:
                raise  # not EEXIST
        return open(os.path.join(self._base_dir, basename), 'wb')



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020-2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Generate tiles from a collection of images with associated WCS coordinate
systems.

This module has the following Python package dependencies:

- astropy
- reproject
- shapely (to optimize the projection in reproject)
"""

__all__ = '''
MultiWcsProcessor
'''.split()

import numpy as np
from tqdm import tqdm
import warnings

from .image import Image, ImageDescription, ImageMode
from .study import StudyTiling

MAXIMUM_CHUNK_SIZE = 128 * 1024 * 1024

class MultiWcsDescriptor(object):
    ident = None
    in_shape = None
    in_wcs = None

    imin = None
    imax = None
    jmin = None
    jmax = None

    chunks = None


class ChunkDescriptor(object):
    j0 = None
    j1 = None
    sub_tiling = None


class MultiWcsProcessor(object):
    def __init__(self, collection):
        self._collection = collection


    def compute_global_pixelization(self, builder):
        from reproject.mosaicking.wcs_helpers import find_optimal_celestial_wcs

        # Load up current WCS information for all of the inputs

        def create_mwcs_descriptor(coll_desc):
            desc = MultiWcsDescriptor()
            desc.ident = coll_desc.collection_id
            desc.in_shape = coll_desc.shape
            desc.in_wcs = coll_desc.wcs
            return desc

        self._descs = [create_mwcs_descriptor(d) for d in self._collection.descriptions()]

        # Compute the optimal tangential tiling that fits all of them. Since WWT
        # tilings must be done in a negative-parity coordinate system, we use an
        # ImageDescription helper to ensure we get that.

        wcs, shape = find_optimal_celestial_wcs(
            ((desc.in_shape, desc.in_wcs) for desc in self._descs),
            auto_rotate = True,
            projection = 'TAN',
        )

        desc = ImageDescription(wcs=wcs, shape=shape)
        desc.ensure_negative_parity()
        self._combined_wcs = desc.wcs
        self._combined_shape = desc.shape
        height, width = self._combined_shape

        self._tiling = StudyTiling(width, height)
        self._tiling.apply_to_imageset(builder.imgset)
        builder.apply_wcs_info(self._combined_wcs, width, height)

        # While we're here, figure out how each input will map onto the global
        # tiling. This makes sure that nothing funky happened during the
        # computation and allows us to know how many tiles we'll have to visit.

        self._n_todo = 0

        for desc in self._descs:
            # XXX: this functionality is largely copied from
            # `reproject.mosaicking.coadd.reproject_and_coadd`, and redundant
            # with it, but it's sufficiently different that I think the best
            # approach is to essentially fork the implementation.

            # Figure out where this array lands in the mosaic.

            ny, nx = desc.in_shape
            xc = np.array([-0.5, nx - 0.5, nx - 0.5, -0.5])
            yc = np.array([-0.5, -0.5, ny - 0.5, ny - 0.5])
            xc_out, yc_out = self._combined_wcs.world_to_pixel(desc.in_wcs.pixel_to_world(xc, yc))

            if np.any(np.isnan(xc_out)) or np.any(np.isnan(yc_out)):
                raise Exception(f'segment {desc.ident} does not fit within the global mosaic')

            desc.imin = max(0, int(np.floor(xc_out.min() + 0.5)))
            desc.imax = min(self._combined_shape[1], int(np.ceil(xc_out.max() + 0.5)))
            desc.jmin = max(0, int(np.floor(yc_out.min() + 0.5)))
            desc.jmax = min(self._combined_shape[0], int(np.ceil(yc_out.max() + 0.5)))

            width = desc.imax - desc.imin
            height = desc.jmax - desc.jmin

            if width <= 0 or height <= 0:
                raise Exception(f'segment {desc.ident} maps to zero size in the global mosaic')

            # Prepare to reproject the image onto the target pixelization.
            # Reproject isn't very memory efficient and can blow up with large
            # images. (My laptop with 64 GB of memory can't handle an
            # all-at-once reprojection of a 12 GB FITS file.) To mitigate this,
            # we reproject each image in chunks, moving down the image
            # vertically in bands. If the image isn't very big this reduces to
            # going all-at-once.

            rows_per_chunk = max(MAXIMUM_CHUNK_SIZE // width, 1)
            desc.chunks = []
            delta_j = 0

            while delta_j < height:
                chunk = ChunkDescriptor()
                chunk.j0 = desc.jmin + delta_j
                chunk.j1 = min(chunk.j0 + rows_per_chunk, desc.jmax)
                chunk_height = chunk.j1 - chunk.j0
                chunk.sub_tiling = self._tiling.compute_for_subimage(
                    desc.imin,
                    chunk.j0,
                    width,
                    chunk_height,
                )
                desc.chunks.append(chunk)

                self._n_todo += chunk.sub_tiling.count_populated_positions()
                delta_j += chunk_height

        return self  # chaining convenience


    def tile(self, pio, reproject_function, parallel=None, cli_progress=False, **kwargs):
        """
        Tile!!!!

        Parameters
        ----------
        pio : :class:`toasty.pyramid.PyramidIO`
            A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
            the tiles in the tile pyramid.
        reproject_function : TKTK
            TKTK
        parallel : integer or None (the default)
            The level of parallelization to use. If unspecified, defaults to using
            all CPUs. If the OS does not support fork-based multiprocessing,
            parallel processing is not possible and serial processing will be
            forced. Pass ``1`` to force serial processing.
        cli_progress : optional boolean, defaults False
            If true, a progress bar will be printed to the terminal using tqdm.

        """
        from .par_util import resolve_parallelism
        parallel = resolve_parallelism(parallel)

        if parallel > 1:
            self._tile_parallel(pio, reproject_function, cli_progress, parallel, **kwargs)
        else:
            self._tile_serial(pio, reproject_function, cli_progress, **kwargs)

        # Since we used `pio.update_image()`, we should clean up the lockfiles
        # that were generated.
        pio.clean_lockfiles(self._tiling._tile_levels)


    def _tile_serial(self, pio, reproject_function, cli_progress, **kwargs):
        invert_into_tiles = pio.get_default_vertical_parity_sign() == 1

        with tqdm(total=self._n_todo, disable=not cli_progress) as progress:
            for image, desc in zip(self._collection.images(), self._descs):
                input_array = image.asarray()

                for chunk in desc.chunks:
                    chunk_shape = (chunk.j1 - chunk.j0, desc.imax - desc.imin)
                    chunk_wcs = self._combined_wcs[chunk.j0:chunk.j1, desc.imin:desc.imax]

                    array = reproject_function(
                        (input_array, image.wcs),
                        output_projection=chunk_wcs,
                        shape_out=chunk_shape,
                        return_footprint=False,
                        **kwargs
                    )

                    image_out = Image.from_array(array.astype(np.float32))

                    for pos, width, height, image_x, image_y, tile_x, tile_y in chunk.sub_tiling.generate_populated_positions():
                        # Because we are doing an arbitrary WCS reprojection anyway,
                        # we can ensure that our source image is stored with a
                        # top-down vertical data layout, AKA negative image parity,
                        # which is what the overall "study" coordinate system needs.
                        # But if we're writing to FITS format tiles, those need to
                        # end up with a bottoms-up format. So we need to flip the
                        # vertical orientation of how we put the data into the tile
                        # buffer.

                        if invert_into_tiles:
                            flip_tile_y1 = 255 - tile_y
                            flip_tile_y0 = flip_tile_y1 - height

                            if flip_tile_y0 == -1:
                                flip_tile_y0 = None  # with a slice, -1 does the wrong thing

                            by_idx = slice(flip_tile_y1, flip_tile_y0, -1)
                        else:
                            by_idx = slice(tile_y, tile_y + height)

                        iy_idx = slice(image_y, image_y + height)
                        ix_idx = slice(image_x, image_x + width)
                        bx_idx = slice(tile_x, tile_x + width)

                        with pio.update_image(pos, masked_mode=image_out.mode, default='masked') as basis:
                            image_out.update_into_maskable_buffer(basis, iy_idx, ix_idx, by_idx, bx_idx)

                        progress.update(1)

        if cli_progress:
            print()


    def _tile_parallel(self, pio, reproject_function, cli_progress, parallel, **kwargs):
        import multiprocessing as mp

        # Start up the workers

        queue = mp.Queue(maxsize = 2 * parallel)
        workers = []

        for _ in range(parallel):
            w = mp.Process(target=_mp_tile_worker, args=(queue, pio, reproject_function, kwargs))
            w.daemon = True
            w.start()
            workers.append(w)

        # Send out them segments

        with tqdm(total=len(self._descs), disable=not cli_progress) as progress:
            for image, desc in zip(self._collection.images(), self._descs):
                queue.put((image, desc, self._combined_wcs))
                progress.update(1)

            queue.close()

            for w in workers:
                w.join()

        if cli_progress:
            print()


def _mp_tile_worker(queue, pio, reproject_function, kwargs):
    """
    Generate and enqueue the tiles that need to be processed.
    """
    from queue import Empty

    invert_into_tiles = pio.get_default_vertical_parity_sign() == 1

    while True:
        try:
            # un-pickling WCS objects always triggers warnings right now
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                image, desc, combined_wcs = queue.get(True, timeout=10)
        except (OSError, ValueError, Empty) as e:
            # OSError or ValueError => queue closed. This signal seems not to
            # cross multiprocess lines, though.
            break

        input_array = image.asarray()

        for chunk in desc.chunks:
            chunk_shape = (chunk.j1 - chunk.j0, desc.imax - desc.imin)
            chunk_wcs = combined_wcs[chunk.j0:chunk.j1, desc.imin:desc.imax]

            array = reproject_function(
                (input_array, image.wcs),
                output_projection=chunk_wcs,
                shape_out=chunk_shape,
                return_footprint=False,
                **kwargs
            )

            image_out = Image.from_array(array.astype(np.float32))

            for pos, width, height, image_x, image_y, tile_x, tile_y in chunk.sub_tiling.generate_populated_positions():
                if invert_into_tiles:
                    flip_tile_y1 = 255 - tile_y
                    flip_tile_y0 = flip_tile_y1 - height

                    if flip_tile_y0 == -1:
                        flip_tile_y0 = None  # with a slice, -1 does the wrong thing

                    by_idx = slice(flip_tile_y1, flip_tile_y0, -1)
                else:
                    by_idx = slice(tile_y, tile_y + height)

                iy_idx = slice(image_y, image_y + height)
                ix_idx = slice(image_x, image_x + width)
                bx_idx = slice(tile_x, tile_x + width)

                with pio.update_image(pos, masked_mode=image_out.mode, default='masked') as basis:
                    image_out.update_into_maskable_buffer(basis, iy_idx, ix_idx, by_idx, bx_idx)



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Azure Blob Storage I/O backend for the pipeline framework.

This module requires that the ``azure.storage.blob`` Python module be available.
If it is not, this module will still be importable, but it won't work. Check the
``ENABLED`` boolean variable or call :func:`assert_enabled` to raise an
exception offering guidance if the needed support is missing.

"""

__all__ = '''
AzureBlobPipelineIo
ENABLED
assert_enabled
'''.split()

import shutil

from . import PipelineIo

try:
    from azure.storage.blob import BlobServiceClient
    ENABLED = True
except ImportError:
    ENABLED = False


def assert_enabled():
    if not ENABLED:
        raise Exception('Azure pipeline I/O backend is needed but unavailable -'
            ' install the `azure-storage-blob` package')


class AzureBlobPipelineIo(PipelineIo):
    """I/O for pipeline processing that uses Microsoft Azure Blob Storage.

    Parameters
    ----------
    connection_string : str
      The Azure "connection string" to use
    container_name : str
      The name of the blob container within the storage account
    path_prefix : str or iterable of str
      A list folder names within the blob container that will be
      prepended to all paths accessed through this object.

    """
    _connection_string = None
    _svc_client = None
    _cnt_client = None
    _container_name = None
    _path_prefix = None

    def __init__(self, connection_string, container_name, path_prefix):
        assert_enabled()

        if isinstance(path_prefix, str):
            path_prefix = (path_prefix, )
        else:
            try:
                path_prefix = tuple(path_prefix)
                for item in path_prefix:
                    assert isinstance(item, str)
            except Exception:
                raise ValueError('path_prefix should be a string or iterable of strings; '
                                 'got %r' % (path_prefix, ))

        self._connection_string = connection_string
        self._container_name = container_name
        self._svc_client = BlobServiceClient.from_connection_string(connection_string)
        self._cnt_client = self._svc_client.get_container_client(container_name)
        self._path_prefix = path_prefix

    def _export_config(self):
        return {
            '_type': 'azure-blob',
            'connection_secret': self._connection_string,
            'container_name': self._container_name,
            'path_prefix': self._path_prefix,
        }

    @classmethod
    def _new_from_config(cls, config):
        return cls(
            config['connection_secret'],
            config['container_name'],
            config['path_prefix'],
        )

    def _make_blob_name(self, path_array):
        """TODO: is this actually correct? Escaping?"""
        return '/'.join(self._path_prefix + tuple(path_array))

    def check_exists(self, *path):
        from azure.core.exceptions import ResourceNotFoundError

        blob_client = self._cnt_client.get_blob_client(self._make_blob_name(path))

        try:
            blob_client.get_blob_properties()
        except ResourceNotFoundError:
            return False
        return True

    def get_item(self, *path, dest=None):
        blob_client = self._cnt_client.get_blob_client(self._make_blob_name(path))
        blob_client.download_blob().readinto(dest)

    def put_item(self, *path, source=None):
        blob_client = self._cnt_client.get_blob_client(self._make_blob_name(path))
        blob_client.upload_blob(source)

    def list_items(self, *path):
        from azure.storage.blob.models import BlobPrefix
        prefix = self._make_blob_name(path) + '/'

        for item in self._cnt_client.list_blobs(
                prefix = prefix,
                delimiter = '/'
        ):
            assert item.name.startswith(prefix)
            stem = item.name[len(prefix):]
            is_folder = isinstance(item, BlobPrefix)

            if is_folder:
                # Returned names end with a '/' too
                assert stem[-1] == '/'
                stem = stem[:-1]

            yield stem, is_folder



from libc.math cimport sin, cos, atan2, hypot
import numpy as np

cimport cython

cimport numpy as np

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

cdef struct Point:
    DTYPE_t x
    DTYPE_t y

cdef void _mid(Point a, Point b, Point *cen):
    """
    Return the midpoint of two points on a great circle arc

    Parameters
    ----------
    a, b: Point
    Two lon, lat pairs, in radians
    """
    cdef DTYPE_t dl, bx, by, b3, l3, outb, outl
    cdef Point out

    dl = b.x - a.x
    bx = cos(b.y) * cos(dl)
    by = cos(b.y) * sin(dl)
    outb = atan2(sin(a.y) + sin(b.y), hypot(cos(a.y) + bx, by))
    outl = a.x + atan2(by, cos(a.y) + bx)
    cen.x = outl
    cen.y = outb


def mid(a, b):
    cdef Point l = Point(a[0], a[1]), m = Point(b[0], b[1]), n
    _mid(l, m, &n)
    return n.x, n.y


@cython.boundscheck(False)
cdef void _subsample(Point ul, Point ur, Point lr, Point ll,
                    DTYPE_t [:, :] x,
                    DTYPE_t [:, :] y,
                    int increasing):
    """
    Given the corners of a toast tile, return the
    sky locations of a subsampled version

    Parameters
    ----------
    ul, ur, lr, ll: Points
        The (lon, lat) coordinates of the four corners of the toast
        tile to subdivide. In radians
    x, y : arrays
        The arrays in which to hold the subdivided locations
    increasing : int
         If 1, the shared edge of the toast tile's two sub-HTM
         trixels increases from left to right. Otherwise, it
         decreases from left to right
    n : int
        The number of pixels in x and y
    """
    cdef Point le, up, lo, ri, cen
    cdef int n = x.shape[0]
    cdef int n2 = n // 2

    _mid(ul, ur, &up)
    _mid(ul, ll, &le)
    _mid(ur, lr, &ri)
    _mid(ll, lr, &lo)

    if increasing:
        _mid(ll, ur, &cen)
    else:
        _mid(ul, lr, &cen)

    if n == 1:
        x[0] = cen.x
        y[0] = cen.y
        return

    _subsample(ul, up, cen, le, x[:n2, :n2], y[:n2, :n2], increasing)
    _subsample(up, ur, ri, cen, x[:n2, n2:], y[:n2, n2:], increasing)
    _subsample(le, cen, lo, ll, x[n2:, :n2], y[n2:, :n2], increasing)
    _subsample(cen, ri, lr, lo, x[n2:, n2:], y[n2:, n2:], increasing)


def subsample(ul, ur, lr, ll, npix, increasing):
    """Subdivide a toast quad, and return the pixel locations

    Parameters
    ----------
    ul, ur, lr, ll : array-like
        Two-element arrays giving the (lon, lat) position of each corner
        of the toast tile, in radians
    npix: int
        The pixel resolution of the subsampled image. Must be a power of 2
    increasing: bool
        Whether the two HTM trixels that define this toast tile are joined
        along the diagonal which increases from left to right


    Returns
    --------
    Two arrays, giving the lon/lat of each point in the subsampled image
    """
    if len(ul) != 2 or len(ur) != 2 or len(lr) != 2 or len(ll) != 2:
        raise ValueError("Toast corners must be two-element arrays")
    if 2 ** int(np.log2(npix)) != npix:
        raise ValueError("npix must be a power of 2: %i" % npix)

    cdef Point _ul, _ur, _lr, _ll
    cdef int _inc = (1 if increasing else 0)

    x = np.zeros((npix, npix), dtype=DTYPE)
    y = np.zeros((npix, npix), dtype=DTYPE)

    _ul = Point(DTYPE(ul[0]), DTYPE(ul[1]))
    _ur = Point(DTYPE(ur[0]), DTYPE(ur[1]))
    _lr = Point(DTYPE(lr[0]), DTYPE(lr[1]))
    _ll = Point(DTYPE(ll[0]), DTYPE(ll[1]))
    _subsample(_ul, _ur, _lr, _ll, x, y, increasing)
    return x, y



# -*- mode: python; coding: utf-8 -*-
# Copyright 2020 the AAS WorldWide Telescope project
# Licensed under the MIT License.

import argparse
import numpy as np
import numpy.testing as nt
import os.path
import pytest

from . import test_path
from .. import cli


class TestMiscCli(object):
    def setup_method(self, method):
        from tempfile import mkdtemp
        self.work_dir = mkdtemp()

    def teardown_method(self, method):
        from shutil import rmtree
        rmtree(self.work_dir)

    def work_path(self, *pieces):
        return os.path.join(self.work_dir, *pieces)

    def test_make_thumbnail(self):
        """
        Just a smoketest.
        """
        args = [
            'make-thumbnail',
            test_path('Equirectangular_projection_SW-tweaked.jpg'),
            self.work_path('basic_cli'),
        ]
        cli.entrypoint(args)

    def test_crop(self):
        """
        Test the generic --crop argument.
        """
        from ..image import ImageLoader

        parser = argparse.ArgumentParser()
        ImageLoader.add_arguments(parser)
        settings = parser.parse_args(['--crop=1,2,3,4'])
        img = ImageLoader.create_from_args(settings).load_path(test_path('crop_input.png'))
        arr = img.asarray()
        assert arr.shape == (256, 256, 3)
        assert arr.max() == 0





# -*- mode: python; coding: utf-8 -*-
# Copyright 2019-2021 the AAS WorldWide Telescope project
# Licensed under the MIT License.

"""
Generate tiles from a collection of images on a common TAN projection.
"""

__all__ = '''
MultiTanProcessor
'''.split()

from astropy.wcs import WCS
import numpy as np
from tqdm import tqdm
import warnings

from .study import StudyTiling

MATCH_HEADERS = [
    'CTYPE1', 'CTYPE2',
    'CRVAL1', 'CRVAL2',
    'CDELT1', 'CDELT2',
]

# In my sample set, some of these files vary minutely from one header to the
# next, so we save them but don't demand exact matching. We should use
# np.isclose() or whatever it is.
SAVE_HEADERS = [
    'PC1_1', 'PC2_2',
]


class MultiTanDescriptor(object):
    ident = None
    in_shape = None

    crxmin = None
    crxmax = None
    crymin = None
    crymax = None

    imin = None
    imax = None
    jmin = None
    jmax = None

    sub_tiling = None


class MultiTanProcessor(object):
    """
    Generate tiles from a collection of images on a common TAN projection.

    Some large astronomical images are stored as a collection of sub-images that
    share a common tangential projection, a format is that is nice and easy to
    convert into a WWT "study" tile pyramid. This class can process a collection
    of such images and break them into the highest-resolution layer of such a
    tile pyramid.
    """

    def __init__(self, collection):
        self._collection = collection


    def compute_global_pixelization(self, builder):
        """
        Read the input images to determine the global pixelation of this data
        set.

        This function reads the FITS headers of all of the input data files to
        determine the overall image size and the parameters of its tiling as a
        WWT study. This should be pretty easy, because we've been assured that
        everything is on a nice common TAN, which is exactly what we need to end
        up with anyway.
        """

        self._descs = []
        ref_headers = None
        global_crxmin = None

        for desc in self._collection.descriptions():
            # For figuring out the tiling properties, we need to ensure that
            # we're working in top-down mode everywhere.
            desc.ensure_negative_parity()

            # Build up information for the common TAN projection.
            header = desc.wcs.to_header()

            if ref_headers is None:
                ref_headers = {}

                for h in MATCH_HEADERS:
                    ref_headers[h] = header[h]

                for h in SAVE_HEADERS:
                    ref_headers[h] = header[h]

                if ref_headers['CTYPE1'] != 'RA---TAN':
                    raise Exception('all inputs must be in a TAN projection, but {} is not'.format(input_id))
                if ref_headers['CTYPE2'] != 'DEC--TAN':
                    raise Exception('all inputs must be in a TAN projection, but {} is not'.format(input_id))
            else:
                for h in MATCH_HEADERS:
                    expected = ref_headers[h]
                    observed = header[h]

                    if observed != expected:
                        raise Exception('inputs are not on uniform WCS grid; in file {}, expected '
                                        'value {} for header {} but observed {}'.format(input_id, expected, h, observed))

            this_crpix1 = header['CRPIX1'] - 1
            this_crpix2 = header['CRPIX2'] - 1

            mtdesc = MultiTanDescriptor()
            mtdesc.ident = desc.collection_id
            mtdesc.in_shape = desc.shape

            mtdesc.crxmin = 0 - this_crpix1
            mtdesc.crxmax = (desc.shape[1] - 1) - this_crpix1
            mtdesc.crymin = 0 - this_crpix2
            mtdesc.crymax = (desc.shape[0] - 1) - this_crpix2

            if global_crxmin is None:
                global_crxmin = mtdesc.crxmin
                global_crxmax = mtdesc.crxmax
                global_crymin = mtdesc.crymin
                global_crymax = mtdesc.crymax
            else:
                global_crxmin = min(global_crxmin, mtdesc.crxmin)
                global_crxmax = max(global_crxmax, mtdesc.crxmax)
                global_crymin = min(global_crymin, mtdesc.crymin)
                global_crymax = max(global_crymax, mtdesc.crymax)

            self._descs.append(mtdesc)

        # We can now compute the global properties of the tiled TAN representation:

        width = int(global_crxmax - global_crxmin) + 1
        height = int(global_crymax - global_crymin) + 1
        self._tiling = StudyTiling(width, height)

        ref_headers['CRPIX1'] = this_crpix1 + 1 + (mtdesc.crxmin - global_crxmin)
        ref_headers['CRPIX2'] = this_crpix2 + 1 + (mtdesc.crymin - global_crymin)
        wcs = WCS(ref_headers)

        self._tiling.apply_to_imageset(builder.imgset)
        builder.apply_wcs_info(wcs, width, height)

        # While we're here, figure out how each input will map onto the global
        # tiling. This makes sure that nothing funky happened during the
        # computation and allows us to know how many tiles we'll have to visit.

        self._n_todo = 0

        for desc in self._descs:
            desc.imin = int(np.floor(desc.crxmin - global_crxmin))
            desc.imax = int(np.ceil(desc.crxmax - global_crxmin))
            desc.jmin = int(np.floor(desc.crymin - global_crymin))
            desc.jmax = int(np.ceil(desc.crymax - global_crymin))

            # Compute the sub-tiling now so that we can count how many total
            # tiles we'll need to process.

            if desc.imax < desc.imin or desc.jmax < desc.jmin:
                raise Exception(f'segment {desc.ident} maps to zero size in the global mosaic')

            desc.sub_tiling = self._tiling.compute_for_subimage(
                desc.imin,
                desc.jmin,
                desc.imax + 1 - desc.imin,
                desc.jmax + 1 - desc.jmin,
            )

            self._n_todo += desc.sub_tiling.count_populated_positions()

        return self  # chaining convenience


    def tile(self, pio, parallel=None, cli_progress=False, **kwargs):
        """
        Tile the input images into the deepest layer of the pyramid.

        Parameters
        ----------
        pio : :class:`toasty.pyramid.PyramidIO`
            A :class:`~toasty.pyramid.PyramidIO` instance to manage the I/O with
            the tiles in the tile pyramid.
        parallel : integer or None (the default)
            The level of parallelization to use. If unspecified, defaults to using
            all CPUs. If the OS does not support fork-based multiprocessing,
            parallel processing is not possible and serial processing will be
            forced. Pass ``1`` to force serial processing.
        cli_progress : optional boolean, defaults False
            If true, a progress bar will be printed to the terminal using tqdm.
        """

        from .par_util import resolve_parallelism
        parallel = resolve_parallelism(parallel)

        if parallel > 1:
            self._tile_parallel(pio, cli_progress, parallel, **kwargs)
        else:
            self._tile_serial(pio, cli_progress, **kwargs)

        # Since we used `pio.update_image()`, we should clean up the lockfiles
        # that were generated.
        pio.clean_lockfiles(self._tiling._tile_levels)


    def _tile_serial(self, pio, cli_progress, **kwargs):
        tile_parity_sign = pio.get_default_vertical_parity_sign()

        with tqdm(total=self._n_todo, disable=not cli_progress) as progress:
            for image, desc in zip(self._collection.images(), self._descs):
                # Ensure that the image and the eventual tile agree on parity
                if image.get_parity_sign() != tile_parity_sign:
                    image.flip_parity()

                for pos, width, height, image_x, image_y, tile_x, tile_y in desc.sub_tiling.generate_populated_positions():
                    # Tiling coordinate systems are always negative (top-down)
                    # parity, with the Y=0 tile at the top. But the actual data
                    # buffers might be positive (bottoms-up) parity -- this is
                    # the case for FITS. In those situations, we have to flip
                    # the vertical positioning. Because we have ensured that the
                    # source image and tile layouts agree, the needed tweak is
                    # actually quite minimal:

                    if tile_parity_sign == 1:
                        image_y = image.height - (image_y + height)
                        tile_y = 256 - (tile_y + height)

                    ix_idx = slice(image_x, image_x + width)
                    bx_idx = slice(tile_x, tile_x + width)
                    iy_idx = slice(image_y, image_y + height)
                    by_idx = slice(tile_y, tile_y + height)

                    with pio.update_image(pos, masked_mode=image.mode, default='masked') as basis:
                        image.update_into_maskable_buffer(basis, iy_idx, ix_idx, by_idx, bx_idx)

                    progress.update(1)

        if cli_progress:
            print()


    def _tile_parallel(self, pio, cli_progress, parallel, **kwargs):
        import multiprocessing as mp

        # Start up the workers

        queue = mp.Queue(maxsize = 2 * parallel)
        workers = []

        for _ in range(parallel):
            w = mp.Process(target=_mp_tile_worker, args=(queue, pio, kwargs))
            w.daemon = True
            w.start()
            workers.append(w)

        # Send out them segments

        with tqdm(total=len(self._descs), disable=not cli_progress) as progress:
            for image, desc in zip(self._collection.images(), self._descs):
                queue.put((image, desc))
                progress.update(1)

            queue.close()

            for w in workers:
                w.join()

        if cli_progress:
            print()


def _mp_tile_worker(queue, pio, kwargs):
    """
    Generate and enqueue the tiles that need to be processed.
    """
    from queue import Empty

    tile_parity_sign = pio.get_default_vertical_parity_sign()

    while True:
        try:
            # un-pickling WCS objects always triggers warnings right now
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                image, desc = queue.get(True, timeout=1)
        except (OSError, ValueError, Empty):
            # OSError or ValueError => queue closed. This signal seems not to
            # cross multiprocess lines, though.
            break

        if image.get_parity_sign() != tile_parity_sign:
            image.flip_parity()

        for pos, width, height, image_x, image_y, tile_x, tile_y in desc.sub_tiling.generate_populated_positions():
            if tile_parity_sign == 1:
                image_y = image.height - (image_y + height)
                tile_y = 256 - (tile_y + height)

            ix_idx = slice(image_x, image_x + width)
            bx_idx = slice(tile_x, tile_x + width)
            iy_idx = slice(image_y, image_y + height)
            by_idx = slice(tile_y, tile_y + height)

            with pio.update_image(pos, masked_mode=image.mode, default='masked') as basis:
                image.update_into_maskable_buffer(basis, iy_idx, ix_idx, by_idx, bx_idx)
